@@ -2,11 +2,7 @@ from enum import Enum
 
 from palimpzest.query.execution.all_sample_execution_strategy import AllSamplingExecutionStrategy
 from palimpzest.query.execution.mab_execution_strategy import MABExecutionStrategy
-from palimpzest.query.execution.parallel_execution_strategy import (
-    ParallelExecutionStrategy,
-    SequentialParallelExecutionStrategy,
-)
-from palimpzest.query.execution.random_sampling_execution_strategy import RandomSamplingExecutionStrategy
+from palimpzest.query.execution.parallel_execution_strategy import ParallelExecutionStrategy
 from palimpzest.query.execution.single_threaded_execution_strategy import (
     PipelinedSingleThreadExecutionStrategy,
     SequentialSingleThreadExecutionStrategy,
@@ -18,9 +14,11 @@ class ExecutionStrategyType(Enum):
     SEQUENTIAL = SequentialSingleThreadExecutionStrategy
     PIPELINED = PipelinedSingleThreadExecutionStrategy
     PARALLEL = ParallelExecutionStrategy
-    SEQUENTIAL_PARALLEL = SequentialParallelExecutionStrategy
+
+    def is_fully_parallel(self) -> bool:
+        """Check if the execution strategy executes operators in parallel."""
+        return self == ExecutionStrategyType.PARALLEL
 
 class SentinelExecutionStrategyType(Enum):
     MAB = MABExecutionStrategy
-    RANDOM = RandomSamplingExecutionStrategy
     ALL = AllSamplingExecutionStrategy
