@@ -3,8 +3,10 @@ from __future__ import annotations
 import logging
 from typing import Any
 
-from palimpzest.core.data.dataclasses import PlanCost
+from palimpzest.core.models import PlanCost
 from palimpzest.policy import Policy
+from palimpzest.query.execution.execution_strategy_type import ExecutionStrategyType
+from palimpzest.query.operators.join import JoinOp
 from palimpzest.query.optimizer.cost_model import BaseCostModel
 from palimpzest.query.optimizer.optimizer_strategy_type import OptimizationStrategyType
 from palimpzest.query.optimizer.primitives import Expression, Group
@@ -64,17 +66,19 @@ class OptimizeGroup(Task):
             task = OptimizePhysicalExpression(physical_expr)
             new_tasks.append(task)
 
+        # and first explore the group if it hasn't been explored yet
+        if not group.explored:
+            task = ExploreGroup(self.group_id)
+            new_tasks.append(task)
+
         logger.debug(f"Done optimizing group {self.group_id}")
         logger.debug(f"New tasks: {len(new_tasks)}")
         return new_tasks
 
 
-class ExpandGroup(Task):
+class ExploreGroup(Task):
     """
-    The task to expand a group.
-
-    NOTE: we currently do not use this task, but I'm keeping it around in case we need it
-    once we add join operations.
+    The task to explore a group and add additional logical expressions.
     """
 
     def __init__(self, group_id: int):
@@ -98,6 +102,12 @@ class ExpandGroup(Task):
             task = OptimizeLogicalExpression(logical_expr, exploring=True)
             new_tasks.append(task)
 
+        # but first (tasks are LIFO), we recursively explore input groups of logical expressions in this group
+        for logical_expr in group.logical_expressions:
+            for input_group_id in logical_expr.input_group_ids:
+                task = ExploreGroup(input_group_id)
+                new_tasks.append(task)
+
         # mark the group as explored and return tasks
         group.set_explored()
 
@@ -120,15 +130,20 @@ class OptimizeLogicalExpression(Task):
 
     def perform(
         self,
-        transformation_rules: list[TransformationRule],
-        implementation_rules: list[ImplementationRule],
+        transformation_rules: list[type[TransformationRule]],
+        implementation_rules: list[type[ImplementationRule]],
         context: dict[str, Any] | None = None,
     ) -> list[Task]:
         logger.debug(f"Optimizing logical expression {self.logical_expression}")
-        # if we're exploring, only apply transformation rules
         if context is None:
             context = {}
-        rules = transformation_rules if self.exploring else transformation_rules + implementation_rules
+
+        # if we're exploring, only apply transformation rules
+        rules = (
+            [rule for rule in transformation_rules if rule.is_exploration_rule()]
+            if self.exploring
+            else transformation_rules + implementation_rules
+        )
 
         # filter out rules that have already been applied to logical expression
         rules = list(filter(lambda rule: rule.get_rule_id() not in self.logical_expression.rules_applied, rules))
@@ -170,7 +185,7 @@ class ApplyRule(Task):
     - schedule OptimizePhysicalExpression tasks
     """
 
-    def __init__(self, rule: Rule, logical_expression: Expression, exploring: bool = False):
+    def __init__(self, rule: type[Rule], logical_expression: Expression, exploring: bool = False):
         self.rule = rule
         self.logical_expression = logical_expression
         self.exploring = exploring
@@ -183,16 +198,13 @@ class ApplyRule(Task):
         **physical_op_params,
     ) -> tuple[list[Task], int]:
         logger.debug(f"Applying rule {self.rule} to logical expression {self.logical_expression}")
-        
-        # check if rule has already been applied to this logical expression; return [] if so
         if context is None:
             context = {}
+
+        # check if rule has already been applied to this logical expression; return [] if so
         if self.rule.get_rule_id() in self.logical_expression.rules_applied:
             return []
 
-        # MAYBE ?TODO?: iterate over bindings for logical expression and rule?
-        #               perhaps some rules can be applied more than once to an expression?
-
         # get the group of the logical expression
         group_id = self.logical_expression.group_id
         group = groups[group_id]
@@ -206,8 +218,8 @@ class ApplyRule(Task):
             )
 
             # filter out any expressions which are duplicates (i.e. they've been previously computed)
-            new_expressions = [expr for expr in new_expressions if expr.get_expr_id() not in expressions]
-            expressions.update({expr.get_expr_id(): expr for expr in new_expressions})
+            new_expressions = [expr for expr in new_expressions if expr.expr_id not in expressions]
+            expressions.update({expr.expr_id: expr for expr in new_expressions})
 
             # add all new groups to the groups mapping
             for group in new_groups:
@@ -234,11 +246,19 @@ class ApplyRule(Task):
         else:
             # apply implementation rule
             new_expressions = self.rule.substitute(self.logical_expression, **physical_op_params)
-            new_expressions = [expr for expr in new_expressions if expr.get_expr_id() not in expressions]
+            new_expressions = [expr for expr in new_expressions if expr.expr_id not in expressions]
+
+            # get the costed_full_op_ids from the context (if provided) and compute whether this
+            # logical expression has physical operators which have been costed
             costed_full_op_ids = context['costed_full_op_ids']
-            if costed_full_op_ids is not None:
+            logical_op_has_been_costed = costed_full_op_ids is not None and any([
+                op_id.split("-")[0] == self.logical_expression.operator.get_logical_op_id()
+                for op_id in costed_full_op_ids
+            ])
+
+            if logical_op_has_been_costed:
                 new_expressions = [expr for expr in new_expressions if expr.operator.get_full_op_id() in costed_full_op_ids]
-            expressions.update({expr.get_expr_id(): expr for expr in new_expressions})
+            expressions.update({expr.expr_id: expr for expr in new_expressions})
             group.physical_expressions.update(new_expressions)
 
             # create new task
@@ -412,8 +432,9 @@ class OptimizePhysicalExpression(Task):
         if context is None:
             context = {}
 
-        # get the optimizer strategy (type) from the context
+        # get the optimizer strategy (type) and the execution strategy (type) from the context
         optimizer_strategy: OptimizationStrategyType = context['optimizer_strategy']
+        execution_strategy: ExecutionStrategyType = context['execution_strategy']
 
         # return if we've already computed the cost of this physical expression
         if optimizer_strategy.is_pareto() and self.physical_expression.pareto_optimal_plan_costs is not None:
@@ -422,57 +443,90 @@ class OptimizePhysicalExpression(Task):
         if optimizer_strategy.is_not_pareto() and self.physical_expression.plan_cost is not None:
             return []
 
-        # for expressions with an input group, compute the input plan cost(s)
-        best_input_plan_cost = PlanCost(cost=0, time=0, quality=1)
-        input_plan_costs = [PlanCost(cost=0, time=0, quality=1)]
+        # for expressions with input group(s), compute the input plan cost(s)
+        best_input_plan_costs = {}
+        pareto_optimal_input_plan_costs = {}
         if len(self.physical_expression.input_group_ids) > 0:
-            # get the input group
-            input_group_id = self.physical_expression.input_group_ids[0]  # TODO: need to handle joins
-            input_group = groups[input_group_id]
-
-            # compute the input plan cost or list of input plan costs
             new_tasks = []
-            if optimizer_strategy.is_not_pareto() and input_group.best_physical_expression is not None:
-                # TODO: apply policy constraint here
-                best_input_plan_cost = input_group.best_physical_expression.plan_cost
+            for input_group_id in self.physical_expression.input_group_ids:
+                # get the input group
+                input_group = groups[input_group_id]
 
-            elif optimizer_strategy.is_pareto() and input_group.pareto_optimal_physical_expressions is not None:
-                # TODO: apply policy constraint here
-                input_plan_costs = []
-                for pareto_physical_expression in input_group.pareto_optimal_physical_expressions:
-                    plan_costs = list(map(lambda tup: tup[0], pareto_physical_expression.pareto_optimal_plan_costs))
-                    input_plan_costs.extend(plan_costs)
+                # compute the input plan cost or list of input plan costs
+                if optimizer_strategy.is_not_pareto() and input_group.best_physical_expression is not None:
+                    # TODO: apply policy constraint here
+                    best_input_plan_costs[input_group_id] = input_group.best_physical_expression.plan_cost
 
-                # NOTE: this list will not necessarily be pareto-optimal, as a plan cost on the pareto frontier of
-                # one pareto_optimal_physical_expression might be dominated by the plan cost on another physical
-                # expression's pareto frontier; we handle this below by taking the pareto frontier of all_possible_plan_costs
-                # de-duplicate equivalent plan costs; we will still reconstruct plans with equivalent cost in optimizer.py
-                input_plan_costs = list(set(input_plan_costs))
+                elif optimizer_strategy.is_pareto() and input_group.pareto_optimal_physical_expressions is not None:
+                    # TODO: apply policy constraint here
+                    input_plan_costs = []
+                    for pareto_physical_expression in input_group.pareto_optimal_physical_expressions:
+                        plan_costs = list(map(lambda tup: tup[0], pareto_physical_expression.pareto_optimal_plan_costs))
+                        input_plan_costs.extend(plan_costs)
 
-            else:
-                task = OptimizeGroup(input_group_id)
-                new_tasks.append(task)
+                    # NOTE: this list will not necessarily be pareto-optimal, as a plan cost on the pareto frontier of
+                    # one pareto_optimal_physical_expression might be dominated by the plan cost on another physical
+                    # expression's pareto frontier; we handle this below by taking the pareto frontier of all_possible_plan_costs
+                    # de-duplicate equivalent plan costs; we will still reconstruct plans with equivalent cost in optimizer.py
+                    pareto_optimal_input_plan_costs[input_group_id] = list(set(input_plan_costs))
+
+                else:
+                    task = OptimizeGroup(input_group_id)
+                    new_tasks.append(task)
 
             # if not all input groups have been costed, we need to compute these first and then retry this task
             if len(new_tasks) > 0:
                 return [self] + new_tasks
 
+        # once all input groups have been costed, compute the cost of this physical expression
         group = groups[self.physical_expression.group_id]
         if optimizer_strategy.is_pareto():
             # compute all possible plan costs for this physical expression given the pareto optimal input plan costs
             all_possible_plan_costs = []
-            for input_plan_cost in input_plan_costs:
-                op_plan_cost = cost_model(self.physical_expression.operator, input_plan_cost.op_estimates)
+            if isinstance(self.physical_expression.operator, JoinOp):
+                assert len(self.physical_expression.input_group_ids) == 2, "Join operator must have exactly two input groups."
 
-                # compute the total cost for this physical expression by summing its operator's PlanCost
-                # with the input groups' total PlanCost; also set the op_estimates for this expression's operator
-                full_plan_cost = op_plan_cost + input_plan_cost
-                full_plan_cost.op_estimates = op_plan_cost.op_estimates
-                all_possible_plan_costs.append((full_plan_cost, input_plan_cost))
+                # get the best input plan costs for both inputs
+                left_input_group_id, right_input_group_id = self.physical_expression.input_group_ids
+                left_best_input_plan_cost = pareto_optimal_input_plan_costs[left_input_group_id]
+                right_best_input_plan_cost = pareto_optimal_input_plan_costs[right_input_group_id]
+                for left_input_plan_cost in left_best_input_plan_cost:
+                    for right_input_plan_cost in right_best_input_plan_cost:
+                        # compute the cost of this operator given the input plan costs
+                        op_plan_cost = cost_model(
+                            self.physical_expression.operator,
+                            left_input_plan_cost.op_estimates,
+                            right_input_plan_cost.op_estimates,
+                        )
+
+                        # compute the total cost for this physical expression by summing its operator's PlanCost
+                        # with the input groups' total PlanCost; also set the op_estimates for this expression's operator
+                        execution_strategy_str = "parallel" if execution_strategy.is_fully_parallel() else "sequential"
+                        full_plan_cost = op_plan_cost.join_add(left_input_plan_cost, right_input_plan_cost, execution_strategy_str)
+                        full_plan_cost.op_estimates = op_plan_cost.op_estimates
+                        all_possible_plan_costs.append((full_plan_cost, (left_input_plan_cost, right_input_plan_cost)))
+
+            else:
+                assert len(self.physical_expression.input_group_ids) < 2, "Non-join operator must have zero or one input groups."
+
+                input_plan_costs = [PlanCost(cost=0, time=0, quality=1)]
+                if len(self.physical_expression.input_group_ids) == 1:
+                    input_group_id = self.physical_expression.input_group_ids[0]
+                    input_plan_costs = pareto_optimal_input_plan_costs[input_group_id]
+
+                # get the pareto-optimal input plan costs for the single input
+                for input_plan_cost in input_plan_costs:
+                    op_plan_cost = cost_model(self.physical_expression.operator, input_plan_cost.op_estimates)
+
+                    # compute the total cost for this physical expression by summing its operator's PlanCost
+                    # with the input groups' total PlanCost; also set the op_estimates for this expression's operator
+                    full_plan_cost = op_plan_cost + input_plan_cost
+                    full_plan_cost.op_estimates = op_plan_cost.op_estimates
+                    all_possible_plan_costs.append((full_plan_cost, (input_plan_cost, None)))
 
             # reduce the set of possible plan costs to the subset which are pareto-optimal
             pareto_optimal_plan_costs = []
-            for idx, (plan_cost, input_plan_cost) in enumerate(all_possible_plan_costs):
+            for idx, (plan_cost, input_plan_cost_tuple) in enumerate(all_possible_plan_costs):
                 pareto_optimal = True
 
                 # check if any other_expr dominates expr
@@ -487,7 +541,7 @@ class OptimizePhysicalExpression(Task):
 
                 # add expr to pareto frontier if it's not dominated
                 if pareto_optimal:
-                    pareto_optimal_plan_costs.append((plan_cost, input_plan_cost))
+                    pareto_optimal_plan_costs.append((plan_cost, input_plan_cost_tuple))
 
             # set the pareto frontier of plan costs which can be obtained by this physical expression
             self.physical_expression.pareto_optimal_plan_costs = pareto_optimal_plan_costs
@@ -496,13 +550,48 @@ class OptimizePhysicalExpression(Task):
             group = self.update_pareto_optimal_physical_expressions(group, policy)
 
         else:
-            # otherwise, compute the cost of this operator given the optimal input plan cost
-            op_plan_cost = cost_model(self.physical_expression.operator, best_input_plan_cost.op_estimates)
 
-            # compute the total cost for this physical expression by summing its operator's PlanCost
-            # with the input groups' total PlanCost; also set the op_estimates for this expression's operator
-            full_plan_cost = op_plan_cost + best_input_plan_cost
-            full_plan_cost.op_estimates = op_plan_cost.op_estimates
+            # otherwise, compute the cost of this operator given the optimal input plan cost(s)
+            full_plan_cost = None
+            if isinstance(self.physical_expression.operator, JoinOp):
+                assert len(self.physical_expression.input_group_ids) == 2, "Join operator must have exactly two input groups."
+
+                # get the best input plan costs for both inputs
+                left_input_group_id, right_input_group_id = self.physical_expression.input_group_ids
+                left_best_input_plan_cost = best_input_plan_costs[left_input_group_id]
+                right_best_input_plan_cost = best_input_plan_costs[right_input_group_id]
+
+                # compute the cost of this operator given the best input plan costs
+                op_plan_cost = cost_model(
+                    self.physical_expression.operator,
+                    left_best_input_plan_cost.op_estimates,
+                    right_best_input_plan_cost.op_estimates,
+                )
+
+                # compute the total cost for this physical expression by summing its operator's PlanCost
+                # with the input groups' total PlanCost; also set the op_estimates for this expression's operator
+                execution_strategy_str = "parallel" if execution_strategy.is_fully_parallel() else "sequential"
+                full_plan_cost = op_plan_cost.join_add(left_best_input_plan_cost, right_best_input_plan_cost, execution_strategy_str)
+                full_plan_cost.op_estimates = op_plan_cost.op_estimates
+
+            else:
+                assert len(self.physical_expression.input_group_ids) < 2, "Non-join operator must have zero or one input groups."
+
+                # get the best input plan cost for the single input
+                best_input_plan_cost = PlanCost(cost=0, time=0, quality=1)
+                if len(self.physical_expression.input_group_ids) == 1:
+                    input_group_id = self.physical_expression.input_group_ids[0]
+                    best_input_plan_cost = best_input_plan_costs[input_group_id]
+
+                # compute the cost of this operator given the best input plan cost
+                op_plan_cost = cost_model(self.physical_expression.operator, best_input_plan_cost.op_estimates)
+
+                # compute the total cost for this physical expression by summing its operator's PlanCost
+                # with the input groups' total PlanCost; also set the op_estimates for this expression's operator
+                full_plan_cost = op_plan_cost + best_input_plan_cost
+                full_plan_cost.op_estimates = op_plan_cost.op_estimates
+
+            # set the plan cost for this physical expression
             self.physical_expression.plan_cost = full_plan_cost
 
             # update the best physical expression for the group
